#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring,too-many-branches
# pylint: disable=too-many-statements,too-many-return-statements,too-few-public-methods
#
# Copyright 2021 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import glob
import sys
from typing import List


def _tokenize(line: str) -> List[str]:

    # remove whitespace
    line = line.strip()
    line = line.replace("\t", "")
    line = line.replace(";", "")

    # find value
    line = line.replace(" ", "|")
    line = line.replace(",", "|")
    line = line.replace(")", "|")
    line = line.replace("(", "|")

    # return empty tokens
    tokens = []
    for token in line.rsplit("|"):
        if token:
            tokens.append(token)
    return tokens


class ReturnValidator:
    def __init__(self):
        self.warnings: List[str] = []

        # internal state
        self._fn = None
        self._line_num = None
        self._value = None
        self._nret = None
        self._rvif = None
        self._line = None

    @property
    def _tokens(self) -> List[str]:
        return _tokenize(self._line)

    @property
    def _value_relaxed(self) -> str:
        if self._value in ["0x0", "0x00", "0x0000"]:
            return "0"
        if self._value in ["0xffffffff"]:
            return "G_MAXUINT32"
        if self._value in ["0xffff"]:
            return "G_MAXUINT16"
        if self._value in ["0xff"]:
            return "G_MAXUINT8"
        if self._value in ["G_SOURCE_REMOVE"]:
            return "FALSE"
        if self._value in ["G_SOURCE_CONTINUE"]:
            return "TRUE"
        return self._value

    def _test_rvif(self) -> None:

        # parse "g_return_val_if_fail (SOMETHING (foo), NULL);"
        self._value = self._tokens[-1]

        # enumerated enum, so ignore
        if self._value.find("_") != -1:
            return

        # is invalid
        if self._rvif and self._value_relaxed not in self._rvif:
            self.warnings.append(
                "{} line {} got {}, expected {}".format(
                    self._fn, self._line_num, self._value, ", ".join(self._rvif)
                )
            )

    def _test_return(self) -> None:

        # parse "return 0x0;"
        self._value = self._tokens[-1]

        # is invalid
        if self._nret and self._value_relaxed in self._nret:
            self.warnings.append(
                "{} line {} got {}, which is not valid".format(
                    self._fn, self._line_num, self._value
                )
            )

    def parse(self, fn: str) -> None:

        self._fn = fn
        with open(fn) as f:
            self._rvif = None
            self._nret = None
            self._line_num = 0
            for line in f.readlines():
                self._line_num += 1
                line = line.rstrip()
                if not line:
                    continue
                if line.endswith("\\"):
                    continue
                if line.endswith("&&"):
                    continue
                self._line = line
                idx = line.find("g_return_val_if_fail")
                if idx != -1:
                    self._test_rvif()
                    continue
                idx = line.find("return")
                if idx != -1:
                    # continue
                    if len(self._tokens) == 2:
                        self._test_return()
                    continue

                # not a function header
                if line[0] in ["#", " ", "\t", "{", "}", "/"]:
                    continue

                # label
                if line.endswith(":"):
                    continue

                # remove prefixes
                if line.startswith("static"):
                    line = line[7:]
                if line.startswith("inline"):
                    line = line[7:]

                # a pointer
                if line.endswith("*"):
                    self._rvif = ["NULL"]
                    self._nret = ["FALSE"]
                    continue

                # not a leading line
                if line.find(" ") != -1:
                    continue

                # a type we know
                if line in ["void"]:
                    self._rvif = []
                    self._nret = []
                    continue
                if line in ["gpointer"]:
                    self._rvif = ["NULL"]
                    self._nret = ["FALSE"]
                    continue
                if line in ["gboolean"]:
                    self._rvif = ["TRUE", "FALSE"]
                    self._nret = ["NULL", "0"]
                    continue
                if line in ["guint32"]:
                    self._rvif = ["0", "G_MAXUINT32"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["GQuark", "GType"]:
                    self._rvif = ["0"]
                    self._nret = ["NULL", "0", "TRUE", "FALSE"]
                    continue
                if line in ["guint64"]:
                    self._rvif = ["0", "G_MAXUINT64"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["guint16"]:
                    self._rvif = ["0", "G_MAXUINT16"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["guint8"]:
                    self._rvif = ["0", "G_MAXUINT8"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gint64"]:
                    self._rvif = ["0", "-1", "G_MAXINT64"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gint32"]:
                    self._rvif = ["0", "-1", "G_MAXINT32"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gint16"]:
                    self._rvif = ["0", "-1", "G_MAXINT16"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gint8"]:
                    self._rvif = ["0", "-1", "G_MAXINT8"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gint", "int"]:
                    self._rvif = ["0", "-1", "G_MAXINT"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["guint"]:
                    self._rvif = ["0", "G_MAXUINT"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gulong"]:
                    self._rvif = ["0", "G_MAXLONG"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gsize", "size_t"]:
                    self._rvif = ["0", "G_MAXSIZE"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                if line in ["gssize", "ssize_t"]:
                    self._rvif = ["0", "-1", "G_MAXSSIZE"]
                    self._nret = ["NULL", "TRUE", "FALSE"]
                    continue
                # print('unknown return type {}'.format(line))
                self._rvif = None
                self._nret = None


def test_files():

    # test all C source files
    validator = ReturnValidator()
    for fn in glob.glob("**/*.c", recursive=True):
        validator.parse(fn)
    for warning in validator.warnings:
        print(warning)

    return 1 if validator.warnings else 0


if __name__ == "__main__":

    # all done!
    sys.exit(test_files())



#!/usr/bin/python3
#
# Copyright 2017 Dell Inc.
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#

import os
import sys
import subprocess
import argparse

CLANG_DIFF_FORMATTERS = [
    "clang-format-diff-11",
    "clang-format-diff-13",
    "clang-format-diff",
]


def parse_args():
    parser = argparse.ArgumentParser(
        description="Reformat C code to match project style",
        epilog="Call with no argument to reformat uncommitted code.",
    )
    parser.add_argument(
        "commit", nargs="*", default="", help="Reformat all changes since this commit"
    )
    parser.add_argument(
        "--debug", action="store_true", help="Display all launched commands"
    )
    return parser.parse_args()


def select_clang_version(formatters):
    for formatter in formatters:
        try:
            ret = subprocess.check_call(
                [formatter, "--help"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            if ret == 0:
                return formatter
        except FileNotFoundError:
            continue
    print("No clang formatter installed")
    sys.exit(1)


## Entry Point ##
if __name__ == "__main__":
    args = parse_args()
    base = os.getenv("GITHUB_BASE_REF")
    if base:
        base = "origin/%s" % base
    else:
        if args.commit:
            base = args.commit[0]
        else:
            base = "HEAD"
    cmd = ["git", "describe", base]
    if args.debug:
        print(cmd)
    ret = subprocess.run(cmd, capture_output=True)
    if ret.returncode:
        if args.debug:
            print(ret.stderr)
        base = "HEAD"
    print("Reformatting code against %s" % base)
    formatter = select_clang_version(CLANG_DIFF_FORMATTERS)
    cmd = ["git", "diff", "-U0", base]
    if args.debug:
        print(cmd)
    ret = subprocess.run(cmd, capture_output=True, text=True)
    if ret.returncode:
        print("Failed to run %s\n%s" % (cmd, ret.stderr.strip()))
        sys.exit(1)
    cmd = [formatter, "-p1"]
    if args.debug:
        print(cmd)
    ret = subprocess.run(cmd, input=ret.stdout, capture_output=True, text=True)
    if ret.returncode:
        print("Failed to run %s\n%s" % (cmd, ret.stderr.strip()))
        sys.exit(1)
    cmd = ["patch", "-p0"]
    if args.debug:
        print(cmd)
    ret = subprocess.run(cmd, input=ret.stdout, capture_output=True, text=True)
    if ret.returncode:
        print("Failed to run %s\n%s" % (cmd, ret.stderr.strip()))
        sys.exit(1)
    sys.exit(0)



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright 2017 Richard Hughes <richard@hughsie.com>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import sys
import argparse
import xml.etree.ElementTree as ET

XMLNS = "{http://www.gtk.org/introspection/core/1.0}"
XMLNS_C = "{http://www.gtk.org/introspection/c/1.0}"


def parse_version(ver):
    return tuple(map(int, ver.split(".")))


def usage(return_code):
    """print usage and exit with the supplied return code"""
    if return_code == 0:
        out = sys.stdout
    else:
        out = sys.stderr
    out.write("usage: %s <NAME> <INPUT> <OUTPUT>\n" % sys.argv[0])
    sys.exit(return_code)


class LdVersionScript:
    """Rasterize some text"""

    def __init__(self, library_name):
        self.library_name = library_name
        self.releases = {}
        self.overrides = {}

    def _add_node(self, node):
        identifier = node.attrib[XMLNS_C + "identifier"]
        introspectable = int(node.get("introspectable", 1))
        version = node.get("version", None)
        if introspectable and not version:
            print("No version for", identifier)
            sys.exit(1)
        if not version:
            return None
        version = node.attrib["version"]
        if version not in self.releases:
            self.releases[version] = []
        release = self.releases[version]
        if identifier not in release:
            release.append(identifier)
        return version

    def _add_cls(self, cls):

        # add all class functions
        for node in cls.findall(XMLNS + "function"):
            self._add_node(node)

        # choose the lowest version method for the _get_type symbol
        version_lowest = None

        # add all class methods
        for node in cls.findall(XMLNS + "method"):
            version_tmp = self._add_node(node)
            if version_tmp:
                if not version_lowest or parse_version(version_tmp) < parse_version(
                    version_lowest
                ):
                    version_lowest = version_tmp

        # add the constructor
        for node in cls.findall(XMLNS + "constructor"):
            version_tmp = self._add_node(node)
            if version_tmp:
                if not version_lowest or parse_version(version_tmp) < parse_version(
                    version_lowest
                ):
                    version_lowest = version_tmp

        if "{http://www.gtk.org/introspection/glib/1.0}get-type" not in cls.attrib:
            return
        type_name = cls.attrib["{http://www.gtk.org/introspection/glib/1.0}get-type"]

        # finally add the get_type symbol
        version = self.overrides.get(type_name, version_lowest)
        if version:
            self.releases[version].append(type_name)

    def import_gir(self, filename):
        tree = ET.parse(filename)
        root = tree.getroot()
        for ns in root.findall(XMLNS + "namespace"):
            for node in ns.findall(XMLNS + "function"):
                self._add_node(node)
            for cls in ns.findall(XMLNS + "record"):
                self._add_cls(cls)
            for cls in ns.findall(XMLNS + "class"):
                self._add_cls(cls)

    def render(self):

        # get a sorted list of all the versions
        versions = []
        for version in self.releases:
            versions.append(version)

        # output the version data to a file
        verout = "# generated automatically, do not edit!\n"
        oldversion = None
        for version in sorted(versions, key=parse_version):
            symbols = sorted(self.releases[version])
            verout += "\n%s_%s {\n" % (self.library_name, version)
            verout += "  global:\n"
            for symbol in symbols:
                verout += "    %s;\n" % symbol
            verout += "  local: *;\n"
            if oldversion:
                verout += "} %s_%s;\n" % (self.library_name, oldversion)
            else:
                verout += "};\n"
            oldversion = version
        return verout


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-r", "--override", action="append", nargs=2, metavar=("symbol", "version")
    )
    args, argv = parser.parse_known_args()
    if len(argv) != 3:
        usage(1)

    ld = LdVersionScript(library_name=argv[0])
    if args.override:
        for override_symbol, override_version in args.override:
            ld.overrides[override_symbol] = override_version
    ld.import_gir(argv[1])
    open(argv[2], "w", newline="\n").write(ld.render())



#!/usr/bin/python3
# pylint: disable=invalid-name,missing-docstring
#
# Copyright 2023 Chun-wei Fan <fanc999@yahoo.com.tw>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

# Script to generate .def file from .map files

import os
import re
import sys


def get_sym_groups(lines, namespace):
    sym_groups = []
    search_header = True
    collect_symbols = False
    search_tail = False

    # namespace + version of library
    namespace_with_ver_regex = re.escape(namespace) + r"_[0-9]+\.[0-9]+\.[0-9]+"
    # Look for symbols added in version in library from map file,
    # which defines a group of symbols
    header_regex = re.compile(r"^" + namespace_with_ver_regex)
    # Look for where to start collecting symbols for the group
    global_regex = re.compile(r"^\s+global:\s+$")
    # Look for where to end collecting symbols for the group
    local_regex = re.compile(r"^\s+local:\s+\*+;\s+$")
    # Look, if any, since which version were the symbols
    # group added
    ending_notail_regex = re.compile(r"^};\s+$")
    ending_tail_regex = re.compile(r"^}\s+" + namespace_with_ver_regex + r";\s+$")

    for l in lines:
        # New symbols group found in map file
        if search_header and header_regex.match(l):
            symbols = []
            header = l.split()[0]
            search_header = False

        # Go through map file to gather up info for this symbols group
        elif not search_header:
            if not collect_symbols and global_regex.match(l):
                collect_symbols = True
            elif collect_symbols:
                tail = None

                # Stop gathering for group if `local: *;` is found
                if local_regex.match(l):
                    search_tail = True
                elif search_tail:
                    # Populate symbols group info and append
                    # to the groups we gathered previously
                    if ending_notail_regex.match(l) or ending_tail_regex.match(l):
                        if ending_tail_regex.match(l):
                            tail_ = l.split()[1]
                            tail = tail_[: tail_.index(";")]
                        sym_group = {
                            "header": header,
                            "symbols": symbols,
                            "tail": tail,
                        }
                        sym_groups.append(sym_group)
                        search_tail = False
                        collect_symbols = False
                        search_header = True

                # Gather up symbols for this group
                else:
                    symbol = l[: l.index(";")].strip()
                    symbols.append(symbol)

    return sym_groups


def output_def_file(def_file, sym_groups):
    with open(def_file, "w") as out:
        out.write("EXPORTS\n")
        for group in sym_groups:
            out.write("\n")
            if group["tail"] is not None:
                out.write(
                    "; APIs added in %s since %s\n" % (group["header"], group["tail"])
                )
            else:
                out.write("; APIs added in %s\n" % group["header"])
            for symbol in group["symbols"]:
                out.write("%s\n" % symbol)


if __name__ == "__main__":
    if len(sys.argv) != 4:
        raise SystemExit("%s <mapfile> <namespace> <deffile>" % sys.argv[0])
    mapfile = sys.argv[1]
    namespace = sys.argv[2]
    def_file = sys.argv[3]
    with open(mapfile, "r") as m:
        lines = m.readlines()
        groups = get_sym_groups(lines, namespace)
        output_def_file(def_file, groups)
